#pragma once

// Includes:

// Std.
#include <vector>
#include <iostream>
// GL
#include <GL/glew.h>
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>


// Defines several possible options for camera movement. Used as abstraction to stay away from window-system specific input methods
enum Camera_Movement {
	FORWARD,
	BACKWARD,
	LEFT,
	RIGHT
};

// Default camera values
const GLfloat YAW = 140.0f;
const GLfloat PITCH = 0.0f;
const GLfloat SPEED = 3.0f;
const GLfloat SENSITIVTY = 0.25f;
const GLfloat ZOOM = 1.0f;

// An abstract camera class that processes input and calculates the corresponding Eular Angles, Vectors and Matrices for use in OpenGL
class Camera
{
public:
	GLboolean GODMODE;
	// Camera Attributes
	glm::vec3 Position;
	glm::vec3 Front;
	glm::vec3 Up;
	glm::vec3 Right;
	glm::vec3 WorldUp;
	// Eular Angles
	GLfloat Yaw;
	GLfloat Pitch;
	// Camera options
	GLfloat MovementSpeed;
	GLfloat MouseSensitivity;
	GLfloat Zoom;
	// For temporary gameplay
	bool actionModelCollected = false;
	bool bearCollected = false;
	bool rulerCollected = false;
	bool planeCollected = false;
	bool cubeCollected = false;
	bool carCollected = false;
	bool jeepCollected = false;
	bool candy1Collected = false;
	bool candy2Collected = false;
	bool toyplaneCollected = false;
	bool squirrelCollected = false;
	bool dogCollected = false;
	bool legoCollected = false;
	bool diggerCollected = false;

	int collectedItemCount = 0;
	int collectableItemsCount = 10;
	bool won = false;
	bool lost = false;

	// Constructor with vectors
	Camera(GLboolean godmode, glm::vec3 position = glm::vec3(0.0f, 0.0f, 0.0f), glm::vec3 up = glm::vec3(0.0f, 1.0f, 0.0f), GLfloat yaw = YAW, GLfloat pitch = PITCH) : Front(glm::vec3(0.0f, 0.0f, -1.0f)), MovementSpeed(SPEED), MouseSensitivity(SENSITIVTY), Zoom(ZOOM)
	{
		this->GODMODE = godmode;
		this->Position = position;
		this->WorldUp = up;
		this->Yaw = yaw;
		this->Pitch = pitch;
		this->updateCameraVectors();
	}
	// Constructor with scalar values
	Camera(GLfloat posX, GLfloat posY, GLfloat posZ, GLfloat upX, GLfloat upY, GLfloat upZ, GLfloat yaw, GLfloat pitch) : Front(glm::vec3(0.0f, 0.0f, -1.0f)), MovementSpeed(SPEED), MouseSensitivity(SENSITIVTY), Zoom(ZOOM)
	{
		this->Position = glm::vec3(posX, posY, posZ);
		this->WorldUp = glm::vec3(upX, upY, upZ);
		this->Yaw = yaw;
		this->Pitch = pitch;
		this->updateCameraVectors();
	}

	// Returns the view matrix calculated using Eular Angles and the LookAt Matrix
	glm::mat4 GetViewMatrix()
	{
		return glm::lookAt(this->Position, this->Position + this->Front, this->Up);
	}

	// Processes input received from any keyboard-like input system. Accepts input parameter in the form of camera defined ENUM (to abstract it from windowing systems)
	void ProcessKeyboard(Camera_Movement direction, GLfloat deltaTime)
	{
		GLfloat velocity = this->MovementSpeed * deltaTime;
		if (direction == FORWARD)
			this->Position += this->Front * velocity;
		if (direction == BACKWARD)
			this->Position -= this->Front * velocity;
		if (direction == LEFT)
			this->Position -= this->Right * velocity;
		if (direction == RIGHT)
			this->Position += this->Right * velocity;
	}

	// Processes input received from a mouse input system. Expects the offset value in both the x and y direction.
	void ProcessMouseMovement(GLfloat xoffset, GLfloat yoffset, GLboolean constrainPitch = true)
	{
		xoffset *= this->MouseSensitivity;
		yoffset *= this->MouseSensitivity;


		if (!GODMODE) {
			GLfloat maxValue = 2.0f;

			//xoffset = xoffset > maxValue ? maxValue : xoffset;
			//xoffset = xoffset < -maxValue ? -maxValue : xoffset;

			yoffset = yoffset > maxValue ? maxValue : yoffset;
			yoffset = yoffset < -maxValue ? -maxValue : yoffset;
		}
		
		this->Yaw += xoffset;
		this->Pitch += yoffset;

		// Make sure that when pitch is out of bounds, screen doesn't get flipped
		if (constrainPitch)
		{
			if (this->Pitch > 89.0f)
				this->Pitch = 89.0f;
			if (this->Pitch < -89.0f)
				this->Pitch = -89.0f;
		}

		// Update Front, Right and Up Vectors using the updated Eular angles
		this->updateCameraVectors();
	}

	// Processes input received from a mouse scroll-wheel event. Only requires input on the vertical wheel-axis
	void ProcessMouseScroll(GLfloat yoffset)
	{
		if (this->Zoom >= 1.0f && this->Zoom <= 45.0f)
			this->Zoom -= yoffset;
		if (this->Zoom <= 1.0f)
			this->Zoom = 1.0f;
		if (this->Zoom >= 45.0f)
			this->Zoom = 45.0f;
	}


	void collectAndWinCondition(glm::vec3 &bearPosition, glm::vec3 &rulerPosition, glm::vec3 &carPosition, glm::vec3 &jeepPosition, glm::vec3 &candy1Position, 
		glm::vec3 &candy2Position, glm::vec3 &toyplanePosition, glm::vec3 &squirrelPosition, glm::vec3 &dogPosition, glm::vec3 &legoPosition, glm::vec3 &diggerPosition) {
		// comment this line in if you want the camera koordinates to be printed
		// std::cout << "Camera x | y | z: " << Position.x  << " | " << Position.y << " | " << Position.z << " | " << std::endl;
		if (Position.x >= 5.6 || Position.x <= -14.65 || Position.y >= 10.4 || Position.y <= -2.5 || Position.z >= 15.3 || Position.z <= -6) {
			lost = true;
			return;
		}
		if (bearCollected == true && rulerCollected == true && carCollected == true && jeepCollected == true && candy1Collected == true &&
			candy2Collected == true && toyplaneCollected == true && squirrelCollected == true && legoCollected == true && diggerCollected == true) {

			won = true;
			return;
		}

		if(!bearCollected) bearCollected = modelCollected(bearPosition);
		if (!rulerCollected) rulerCollected = modelCollected(rulerPosition);
		if (!carCollected) carCollected = modelCollected(carPosition);
		if (!jeepCollected) jeepCollected = modelCollected(jeepPosition);
		if (!candy1Collected) candy1Collected = modelCollected(candy1Position);
		if (!candy2Collected) candy2Collected = modelCollected(candy2Position);
		if (!toyplaneCollected) toyplaneCollected = modelCollected(toyplanePosition);
		if (!squirrelCollected) squirrelCollected = modelCollected(squirrelPosition);
		//if (!dogCollected) dogCollected = modelCollected(dogPosition);
		if (!legoCollected) legoCollected = modelCollected(legoPosition);
		if (!diggerCollected) diggerCollected = modelCollected(diggerPosition);
	}

	void updateCam() {
		this->updateCameraVectors();
	}

private:
	// Calculates the front vector from the Camera's (updated) Eular Angles
	void updateCameraVectors()
	{
		// Calculate the new Front vector
		glm::vec3 front;
		front.x = cos(glm::radians(this->Yaw)) * cos(glm::radians(this->Pitch));
		front.y = sin(glm::radians(this->Pitch));
		front.z = sin(glm::radians(this->Yaw)) * cos(glm::radians(this->Pitch));
		this->Front = glm::normalize(front);
		// Also re-calculate the Right and Up vector
		this->Right = glm::normalize(glm::cross(this->Front, this->WorldUp));  // Normalize the vectors, because their length gets closer to 0 the more you look up or down which results in slower movement.
		this->Up = glm::normalize(glm::cross(this->Right, this->Front));
	}

	bool modelCollected(glm::vec3 &modelPosition) {
		GLfloat length = glm::length(modelPosition - Position);
		if (length < 1) {
			std::cout << "ITEM COLLECTED! ITEM COLLECTED! ITEM COLLECTED! ITEM COLLECTED! ITEM COLLECTED!" << std::endl;
			collectedItemCount++;
			return true;
		}

		return false;
	}
};


