#pragma once
// Includes:

// Std.
#include <vector>
#include <iostream>
// GL
#include <GL/glew.h>
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>


// Defines several possible options for camera movement. Used as abstraction to stay away from window-system specific input methods
enum Object_Movement {
	FORWARD1,
	BACKWARD1,
	LEFT1,
	RIGHT1
};

// Default object values
const GLfloat YAW1 = -90.0f;
const GLfloat PITCH1 = 0.0f;
const GLfloat SPEED1 = 3.0f;
const GLfloat SENSITIVTY1 = 0.25f;
const GLfloat ZOOM1 = 45.0f;

// An abstract object class that processes input and calculates the corresponding Eular Angles, Vectors and Matrices for use in OpenGL
class MovingObject
{
public:
	// Camera Attributes
	glm::vec3 Position;
	glm::vec3 Front;
	glm::vec3 Up;
	glm::vec3 Right;
	glm::vec3 WorldUp;
	// Eular Angles
	GLfloat Yaw;
	GLfloat Pitch;
	// Camera options
	GLfloat MovementSpeed;
	GLfloat MouseSensitivity;
	GLfloat Zoom;
	// For temporary gameplay
	bool actionModelCollected = false;
	bool bearCollected = false;
	bool rulerCollected = false;
	bool planeCollected = false;
	bool cubeCollected = false;
	bool won = false;
	bool lost = false;

	// Constructor with vectors
	MovingObject(glm::vec3 position = glm::vec3(0.0f, 0.0f, 0.0f), glm::vec3 up = glm::vec3(0.0f, 1.0f, 0.0f), GLfloat yaw = YAW1, GLfloat pitch = PITCH1) : Front(glm::vec3(0.0f, 0.0f, -1.0f)), MovementSpeed(SPEED1), MouseSensitivity(SENSITIVTY1), Zoom(ZOOM1)
	{
		this->Position = position;
		this->WorldUp = up;
		this->Yaw = yaw;
		this->Pitch = pitch;
		this->updateObjectVectors();
	}
	// Constructor with scalar values
	MovingObject(GLfloat posX, GLfloat posY, GLfloat posZ, GLfloat upX, GLfloat upY, GLfloat upZ, GLfloat yaw, GLfloat pitch) : Front(glm::vec3(0.0f, 0.0f, -1.0f)), MovementSpeed(SPEED1), MouseSensitivity(SENSITIVTY1), Zoom(ZOOM1)
	{
		this->Position = glm::vec3(posX, posY, posZ);
		this->WorldUp = glm::vec3(upX, upY, upZ);
		this->Yaw = yaw;
		this->Pitch = pitch;
		this->updateObjectVectors();
	}

	// Returns the view matrix calculated using Eular Angles and the LookAt Matrix
	glm::mat4 GetViewMatrix()
	{
		if (Position.x >= 5.6 || Position.x <= -14.65 || Position.y >= 10.4 || Position.y <= -2.5 || Position.z >= 15.3 || Position.z <= -6) {
			lost = true;
		}
		if (bearCollected == true && rulerCollected == true && planeCollected == true) {
			won = true;
		}
		if (bearCollected == false && Position.x >= -12.0 && Position.x <= -9.0 && Position.y >= -0.8 && Position.y <= 6.2 && Position.z >= -0.5 && Position.z <= 2.5) {
			std::cout << "ITEM COLLECTED! ITEM COLLECTED! ITEM COLLECTED! ITEM COLLECTED! ITEM COLLECTED!" << std::endl;
			bearCollected = true;
		}
		if (rulerCollected == false && Position.x >= -8.0 && Position.x <= -7.0 && Position.y >= 0.5 && Position.y <= 1.5 && Position.z >= -3 && Position.z <= 1.0) {
			std::cout << "ITEM COLLECTED! ITEM COLLECTED! ITEM COLLECTED! ITEM COLLECTED! ITEM COLLECTED!" << std::endl;
			rulerCollected = true;
		}
		if (planeCollected == false && Position.x >= -0.5 && Position.x <= 0.5 && Position.y >= -0.5 && Position.y <= 0.5 && Position.z >= 6.5 && Position.z <= 7.5) {
			std::cout << "ITEM COLLECTED! ITEM COLLECTED! ITEM COLLECTED! ITEM COLLECTED! ITEM COLLECTED!" << std::endl;
			planeCollected = true;
		}

		return glm::lookAt(this->Position, this->Position + this->Front, this->Up);
	}

	// Processes input received from any keyboard-like input system. Accepts input parameter in the form of objects defined ENUM (to abstract it from windowing systems)
	void ProcessKeyboard(Object_Movement direction, GLfloat deltaTime)
	{
		GLfloat velocity = this->MovementSpeed * deltaTime;
		if (direction == FORWARD1)
			this->Position += this->Front * velocity;
		if (direction == BACKWARD1)
			this->Position -= this->Front * velocity;
		if (direction == LEFT1)
			this->Position -= this->Right * velocity;
		if (direction == RIGHT1)
			this->Position += this->Right * velocity;
	}

	// Processes input received from a mouse input system. Expects the offset value in both the x and y direction.
	void ProcessMouseMovement(GLfloat xoffset, GLfloat yoffset, GLboolean constrainPitch = true)
	{
		xoffset *= this->MouseSensitivity;
		yoffset *= this->MouseSensitivity;

		this->Yaw += xoffset;
		this->Pitch += yoffset;

		// Make sure that when pitch is out of bounds, screen doesn't get flipped
		if (constrainPitch)
		{
			if (this->Pitch > 89.0f)
				this->Pitch = 89.0f;
			if (this->Pitch < -89.0f)
				this->Pitch = -89.0f;
		}

		// Update Front, Right and Up Vectors using the updated Eular angles
		this->updateObjectVectors();
	}

	// Processes input received from a mouse scroll-wheel event. Only requires input on the vertical wheel-axis
	void ProcessMouseScroll(GLfloat yoffset)
	{
		if (this->Zoom >= 1.0f && this->Zoom <= 45.0f)
			this->Zoom -= yoffset;
		if (this->Zoom <= 1.0f)
			this->Zoom = 1.0f;
		if (this->Zoom >= 45.0f)
			this->Zoom = 45.0f;
	}

private:
	// Calculates the front vector from the Object's (updated) Eular Angles
	void updateObjectVectors()
	{
		// Calculate the new Front vector
		glm::vec3 front;
		front.x = cos(glm::radians(this->Yaw)) * cos(glm::radians(this->Pitch));
		front.y = sin(glm::radians(this->Pitch));
		front.z = sin(glm::radians(this->Yaw)) * cos(glm::radians(this->Pitch));
		this->Front = glm::normalize(front);
		// Also re-calculate the Right and Up vector
		this->Right = glm::normalize(glm::cross(this->Front, this->WorldUp));  // Normalize the vectors, because their length gets closer to 0 the more you look up or down which results in slower movement.
		this->Up = glm::normalize(glm::cross(this->Right, this->Front));
	}
};

