#include "ParticleGenerator.h"
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>

ParticleGenerator::ParticleGenerator(Shader shader, GLuint amount)
	: shader(shader), amount(amount)
{
	this->init();
}

void ParticleGenerator::Update(GLfloat dt, glm::vec3 position, GLuint newParticles, glm::vec3 offset, GLboolean collected)
{
	// Add new particles 
	for (GLuint i = 0; i < newParticles; ++i)
	{
		int unusedParticle = this->firstUnusedParticle();
		if(!collected) this->respawnParticle(this->particles[unusedParticle], position, offset);
	}
	// Update all particles
	for (GLuint i = 0; i < this->amount; ++i)
	{
		Particle &p = this->particles[i];
		p.Life -= dt*0.3; // reduce life
		if (p.Life > 0.0f)
		{	// particle is alive, thus update
			p.Position += p.Velocity * dt;
			//p.Color.a -= dt * 2.5;
		}
	}
}

// Render all particles
void ParticleGenerator::Draw(glm::mat4 &projection, glm::mat4 &view)
{
	// Use additive blending to give it a 'glow' effect
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	this->shader.Use();
	for (Particle particle : this->particles)
	{
		if (particle.Life > 0.0f)
		{
			this->UpdateModelViewMatrix(particle.Position, particle.Rotation, particle.Scale, view);
			//this->shader.SetVector2f("offset", particle.Position);
			glUniform3f(glGetUniformLocation(shader.Program, "offset"), particle.Position.x, particle.Position.y, particle.Position.z);
			glUniformMatrix4fv(glGetUniformLocation(shader.Program, "projection"), 1, GL_FALSE, glm::value_ptr(projection));
			//glUniform4f(glGetUniformLocation(shader.Program, "color"), particle.Color.x, particle.Color.y, particle.Color.z, particle.Color.w);
			//this->shader.SetVector4f("color", particle.Color);
			//this->texture.Bind();
			glBindVertexArray(this->VAO);
			glDrawArrays(GL_TRIANGLES, 0, 6);
			glBindVertexArray(0);
		}
	}
	// Don't forget to reset to default blending mode
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

void ParticleGenerator::init()
{
	// Set up mesh and attribute properties
	GLuint VBO;
	GLfloat particle_quad[] = {
		0.0f, 1.0f, 0.0f, 1.0f,
		1.0f, 0.0f, 1.0f, 0.0f,
		0.0f, 0.0f, 0.0f, 0.0f,

		0.0f, 1.0f, 0.0f, 1.0f,
		1.0f, 1.0f, 1.0f, 1.0f,
		1.0f, 0.0f, 1.0f, 0.0f
	};
	glGenVertexArrays(1, &this->VAO);
	glGenBuffers(1, &VBO);
	glBindVertexArray(this->VAO);
	// Fill mesh buffer
	glBindBuffer(GL_ARRAY_BUFFER, VBO);
	glBufferData(GL_ARRAY_BUFFER, sizeof(particle_quad), particle_quad, GL_STATIC_DRAW);
	// Set mesh attributes
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 4 * sizeof(GLfloat), (GLvoid*)0);
	glBindVertexArray(0);

	// Create this->amount default particle instances
	for (GLuint i = 0; i < this->amount; ++i)
		this->particles.push_back(Particle());
}

// Stores the index of the last particle used (for quick access to next dead particle)
GLuint lastUsedParticle = 0;
GLuint ParticleGenerator::firstUnusedParticle()
{
	// First search from last used particle, this will usually return almost instantly
	for (GLuint i = lastUsedParticle; i < this->amount; ++i) {
		if (this->particles[i].Life <= 0.0f) {
			lastUsedParticle = i;
			return i;
		}
	}
	// Otherwise, do a linear search
	for (GLuint i = 0; i < lastUsedParticle; ++i) {
		if (this->particles[i].Life <= 0.0f) {
			lastUsedParticle = i;
			return i;
		}
	}
	// All particles are taken, override the first one (note that if it repeatedly hits this case, more particles should be reserved)
	lastUsedParticle = 0;
	return 0;
}

void ParticleGenerator::respawnParticle(Particle &particle, glm::vec3 position, glm::vec3 offset)
{
	GLfloat randomPx = this->RandomFloat(0.0f, 2.0f)-1;
	GLfloat zMax = 2 - glm::abs(randomPx);
	GLfloat randomPz = this->RandomFloat(0.0f, zMax) -(zMax/2);
	GLfloat randomV = this->RandomFloat(0.1f, 0.3f);
	GLfloat randomR = this->RandomFloat(0.0f, 1.0f);
	GLfloat randomS = this->RandomFloat(0.02f, 0.06f);
	GLfloat randomL = this->RandomFloat(1.0f, 1.5f);


	particle.Position = position + offset + glm::vec3(randomPx, 0.0f, randomPz);
	particle.Velocity = glm::vec3(0.0f, 1.0f, 0.0f) * randomV;

	particle.Rotation = randomR;
	particle.Scale = randomS;
	particle.Life = randomL;
	
}

void ParticleGenerator::UpdateModelViewMatrix(glm::vec3 position, float rotation, float scale, glm::mat4 view) {
	glm::mat4 model;
	model = glm::mat4();
	model = glm::translate(model, position);

	model[0][0] = view[0][0];
	model[0][1] = view[1][0];
	model[0][2] = view[2][0];
	model[1][0] = view[0][1];
	model[1][1] = view[1][1];
	model[1][2] = view[2][1];
	model[2][0] = view[0][2];
	model[2][1] = view[1][2];
	model[2][2] = view[2][2];

	model = glm::rotate(model, glm::radians(rotation), glm::vec3(0.0f, 0.0f, 1.0f));
	model = glm::scale(model, glm::vec3(scale));

	glm::mat4 modelView = view * model;
	glUniformMatrix4fv(glGetUniformLocation(shader.Program, "modelView"), 1, GL_FALSE, glm::value_ptr(modelView));
}

GLfloat ParticleGenerator::RandomFloat(GLfloat a, GLfloat b) {
	GLfloat random = ((GLfloat)rand()) / (GLfloat)RAND_MAX;
	GLfloat diff = b - a;
	GLfloat r = random * diff;
	return a + r;
}
